@@ -262,6 +262,7 @@ struct generic_conf {
 	gchar *certfile;        /**< certificate file             */
 	gchar *keyfile;         /**< key file                     */
 	gchar *cacertfile;      /**< CA certificate file          */
+	gchar *tlsprio;		/**< TLS priority string	  */
         gint flags;             /**< global flags                 */
 	gint threads;		/**< maximum number of parallel threads we want to run */
 };
@@ -778,6 +779,7 @@ GArray* parse_cfile(gchar* f, struct generic_conf *const genconf, bool expect_ge
 		{ "certfile",   FALSE, PARAM_STRING,    &(genconftmp.certfile),   0 },
 		{ "keyfile",    FALSE, PARAM_STRING,    &(genconftmp.keyfile),    0 },
 		{ "cacertfile", FALSE, PARAM_STRING,    &(genconftmp.cacertfile), 0 },
+		{ "tlsprio",	FALSE,  PARAM_STRING,   &(genconftmp.tlsprio),    0 },
 	};
 	PARAM* p=gp;
 	int p_size=sizeof(gp)/sizeof(PARAM);
@@ -796,6 +798,8 @@ GArray* parse_cfile(gchar* f, struct generic_conf *const genconf, bool expect_ge
 
         memset(&genconftmp, 0, sizeof(struct generic_conf));
 
+	genconftmp.tlsprio = "NORMAL:-VERS-TLS-ALL:+VERS-TLS1.2:%SERVER_PRECEDENCE";
+
         if (genconf) {
                 /* Use the passed configuration values as defaults. The
                  * parsing algorithm below updates all parameter targets
@@ -1325,8 +1329,7 @@ int expread(off_t a, char *buf, size_t len, CLIENT *client) {
 		if (client->difmap[mapcnt]!=(u32)(-1)) { /* the block is already there */
 			DEBUG("Page %llu is at %lu\n", (unsigned long long)mapcnt,
 			       (unsigned long)(client->difmap[mapcnt]));
-			myseek(client->difffile, client->difmap[mapcnt]*DIFFPAGESIZE+offset);
-			if (read(client->difffile, buf, rdlen) != rdlen) return -1;
+			if (pread(client->difffile, buf, rdlen, client->difmap[mapcnt]*DIFFPAGESIZE+offset) != rdlen) return -1;
 		} else { /* the block is not there */
 			DEBUG("Page %llu is not here, we read the original one\n",
 			       (unsigned long long)mapcnt);
@@ -1371,11 +1374,8 @@ int expwrite(off_t a, char *buf, size_t len, CLIENT *client, int fua) {
 		if (client->difmap[mapcnt]!=(u32)(-1)) { /* the block is already there */
 			DEBUG("Page %llu is at %lu\n", (unsigned long long)mapcnt,
 			       (unsigned long)(client->difmap[mapcnt])) ;
-			myseek(client->difffile,
-					client->difmap[mapcnt]*DIFFPAGESIZE+offset);
-			if (write(client->difffile, buf, wrlen) != wrlen) return -1 ;
+			if (pwrite(client->difffile, buf, wrlen, client->difmap[mapcnt]*DIFFPAGESIZE+offset) != wrlen) return -1 ;
 		} else { /* the block is not there */
-			myseek(client->difffile,client->difffilelen*DIFFPAGESIZE) ;
 			client->difmap[mapcnt]=(client->server->flags&F_SPARSE)?mapcnt:client->difffilelen++;
 			DEBUG("Page %llu is not here, we put it at %lu\n",
 			       (unsigned long long)mapcnt,
@@ -1630,7 +1630,7 @@ CLIENT* handle_starttls(CLIENT* client, int opt, GArray* servers, uint32_t cflag
 	gnutls_certificate_set_verify_function(x509_cred, verify_cert);
 	check_rv(gnutls_certificate_set_x509_trust_file(x509_cred, genconf->cacertfile, GNUTLS_X509_FMT_PEM));
 	check_rv(gnutls_certificate_set_x509_key_file(x509_cred, genconf->certfile, genconf->keyfile, GNUTLS_X509_FMT_PEM));
-	check_rv(gnutls_priority_init(&priority_cache, "NORMAL:-VERS-TLS-ALL:+VERS-TLS1.2:%SERVER_PRECEDENCE", NULL));
+	check_rv(gnutls_priority_init(&priority_cache, genconf->tlsprio, NULL));
 	check_rv(gnutls_init(session, GNUTLS_SERVER));
 	check_rv(gnutls_priority_set(*session, priority_cache));
 	check_rv(gnutls_credentials_set(*session, GNUTLS_CRD_CERTIFICATE, x509_cred));
@@ -2025,7 +2025,7 @@ static void handle_write_zeroes(CLIENT* client, struct nbd_request* req) {
 	// TODO: handle this far more efficiently with reference to the
 	// actual backing driver
 	pthread_mutex_lock(&(client->lock));
-	writeit(client->net, &rep, sizeof rep);
+	socket_write(client, &rep, sizeof rep);
 	pthread_mutex_unlock(&(client->lock));
 }
 
@@ -3242,6 +3242,7 @@ int main(int argc, char *argv[]) {
 	dousers(genconf.user, genconf.group);
 
 #if HAVE_GNUTLS
+	gnutls_global_init();
 	static gnutls_dh_params_t dh_params;
 	gnutls_dh_params_init(&dh_params);
 	gnutls_dh_params_generate2(dh_params,
