@@ -60,6 +60,8 @@ bool address_matches(const char* mask, const struct sockaddr* addr, GError** err
 		uint8_t mask = 0;
 		int len_left = masklen;
 		if(res->ai_family != addr->sa_family) {
+			msg(LOG_DEBUG, "client address does not match %d/%d: address family mismatch (IPv4 vs IPv6?)",
+			    (int)res->ai_family, (int)addr->sa_family);
 			goto next;
 		}
 		switch(addr->sa_family) {
@@ -284,9 +286,3 @@ int exptrim(struct nbd_request* req, CLIENT* client) {
 	DEBUG("Performed TRIM request from %llu to %llu", (unsigned long long) req->from, (unsigned long long) req->len);
 	return 0;
 }
-
-void myseek(int handle,off_t a) {
-	if (lseek(handle, a, SEEK_SET) < 0) {
-		err("Can not seek locally!\n");
-	}
-}
