@@ -42,6 +42,10 @@
 #define MY_NAME "nbd-tester-client"
 #include "cliserv.h"
 
+#if HAVE_GNUTLS
+#include "crypto-gnutls.h"
+#endif
+
 static gchar errstr[1024];
 const static int errstr_len = 1023;
 
@@ -50,6 +54,10 @@ static uint64_t size;
 static int looseordering = 0;
 
 static gchar *transactionlog = "nbd-tester-client.tr";
+static gchar *certfile = NULL;
+static gchar *keyfile = NULL;
+static gchar *cacertfile = NULL;
+static gchar *tlshostname = NULL;
 
 typedef enum {
 	CONNECTION_TYPE_NONE,
@@ -272,6 +280,7 @@ int writebuffer(int fd, struct chunklist *l)
 #define TEST_WRITE (1<<0)
 #define TEST_FLUSH (1<<1)
 #define TEST_EXPECT_ERROR (1<<2)
+#define TEST_HANDSHAKE (1<<3)
 
 int timeval_subtract(struct timeval *result, struct timeval *x,
 		     struct timeval *y)
@@ -341,6 +350,10 @@ static inline int write_all(int f, void *buf, size_t len)
 	return retval;
 }
 
+static int tlserrout (void *opaque, const char *format, va_list ap) {
+	return vfprintf(stderr, format, ap);
+}
+
 #define READ_ALL_ERRCHK(f, buf, len, whereto, errmsg...) if((read_all(f, buf, len))<=0) { snprintf(errstr, errstr_len, ##errmsg); goto whereto; }
 #define READ_ALL_ERR_RT(f, buf, len, whereto, rval, errmsg...) if((read_all(f, buf, len))<=0) { snprintf(errstr, errstr_len, ##errmsg); retval = rval; goto whereto; }
 
@@ -348,7 +361,7 @@ static inline int write_all(int f, void *buf, size_t len)
 #define WRITE_ALL_ERR_RT(f, buf, len, whereto, rval, errmsg...) if((write_all(f, buf, len))<=0) { snprintf(errstr, errstr_len, ##errmsg); retval = rval; goto whereto; }
 
 int setup_connection_common(int sock, char *name, CONNECTION_TYPE ctype,
-			    int *serverflags)
+			    int *serverflags, int testflags)
 {
 	char buf[256];
 	u64 tmp64;
@@ -384,7 +397,12 @@ int setup_connection_common(int sock, char *name, CONNECTION_TYPE ctype,
 		READ_ALL_ERRCHK(sock, &size, sizeof(size), err,
 				"Could not read size: %s", strerror(errno));
 		size = ntohll(size);
-		READ_ALL_ERRCHK(sock, buf, 128, err, "Could not read data: %s",
+		uint32_t flags;
+		READ_ALL_ERRCHK(sock, &flags, sizeof(uint32_t), err,
+				"Could not read flags: %s", strerror(errno));
+		flags = ntohl(flags);
+		*serverflags = flags;
+		READ_ALL_ERRCHK(sock, buf, 124, err, "Could not read data: %s",
 				strerror(errno));
 		goto end;
 	}
@@ -395,9 +413,130 @@ int setup_connection_common(int sock, char *name, CONNECTION_TYPE ctype,
 	/* negotiation flags */
 	if (handshakeflags & NBD_FLAG_FIXED_NEWSTYLE)
 		negotiationflags |= NBD_FLAG_C_FIXED_NEWSTYLE;
+	else if (keyfile) {
+		snprintf(errstr, errstr_len, "Cannot negotiate TLS without NBD_FLAG_FIXED_NEWSTYLE");
+		goto err;
+	}
 	negotiationflags = htonl(negotiationflags);
 	WRITE_ALL_ERRCHK(sock, &negotiationflags, sizeof(negotiationflags), err,
 			 "Could not write reserved field: %s", strerror(errno));
+	if (testflags & TEST_HANDSHAKE) {
+		/* Server must support newstyle for this test */
+		if (!(handshakeflags & NBD_FLAG_FIXED_NEWSTYLE)) {
+			strncpy(errstr, "server does not support handshake", errstr_len);
+			goto err;
+		}
+		goto end;
+	}
+#if HAVE_GNUTLS
+	/* TLS */
+	if (keyfile) {
+		int plainfd[2]; // [0] is used by the proxy, [1] is used by NBD
+		tlssession_t *s = NULL;
+		int ret;
+
+		/* magic */
+		tmp64 = htonll(opts_magic);
+		WRITE_ALL_ERRCHK(sock, &tmp64, sizeof(tmp64), err,
+				 "Could not write magic: %s", strerror(errno));
+		/* starttls */
+		tmp32 = htonl(NBD_OPT_STARTTLS);
+		WRITE_ALL_ERRCHK(sock, &tmp32, sizeof(tmp32), err,
+			 "Could not write option: %s", strerror(errno));
+		/* length of data */
+		tmp32 = htonl(0);
+		WRITE_ALL_ERRCHK(sock, &tmp32, sizeof(tmp32), err,
+			 "Could not write option length: %s", strerror(errno));
+
+		READ_ALL_ERRCHK(sock, &tmp64, sizeof(tmp64), err,
+				"Could not read cliserv_magic: %s", strerror(errno));
+		tmp64 = ntohll(tmp64);
+		if (tmp64 != NBD_OPT_REPLY_MAGIC) {
+			strncpy(errstr, "reply magic does not match", errstr_len);
+			goto err;
+		}
+		READ_ALL_ERRCHK(sock, &tmp32, sizeof(tmp32), err,
+				"Could not read option type: %s", strerror(errno));
+		tmp32 = ntohl(tmp32);
+		if (tmp32 != NBD_OPT_STARTTLS) {
+			strncpy(errstr, "Reply to wrong option", errstr_len);
+			goto err;
+		}
+		READ_ALL_ERRCHK(sock, &tmp32, sizeof(tmp32), err,
+				"Could not read option reply type: %s", strerror(errno));
+		tmp32 = ntohl(tmp32);
+		if (tmp32 != NBD_REP_ACK) {
+			if(tmp32 & NBD_REP_FLAG_ERROR) {
+				snprintf(errstr, errstr_len, "Received error %d", tmp32 & ~NBD_REP_FLAG_ERROR);
+			} else {
+				snprintf(errstr, errstr_len, "Option reply type %d != NBD_REP_ACK", tmp32);
+			}
+			goto err;
+		}
+		READ_ALL_ERRCHK(sock, &tmp32, sizeof(tmp32), err,
+				"Could not read option data length: %s", strerror(errno));
+		tmp32 = ntohl(tmp32);
+		if (tmp32 != 0) {
+			strncpy(errstr, "Option reply data length != 0", errstr_len);
+			goto err;
+		}
+
+		s = tlssession_new(FALSE,
+				   keyfile,
+				   certfile,
+				   cacertfile,
+				   tlshostname,
+				   !cacertfile || !tlshostname, // insecure flag
+#ifdef DODBG
+				   1, // debug
+#else
+				   0, // debug
+#endif
+				   NULL, // quitfn
+				   tlserrout, // erroutfn
+				   NULL // opaque
+			);
+		if (!s) {
+			strncpy(errstr, "Cannot establish TLS session", errstr_len);
+			goto err;
+		}
+
+		if (socketpair(AF_UNIX, SOCK_STREAM, 0, plainfd) < 0) {
+			strncpy(errstr, "Cannot get socket pair", errstr_len);
+			goto err;
+		}
+
+		if (set_nonblocking(plainfd[0], 0) <0 ||
+		    set_nonblocking(plainfd[1], 0) <0 ||
+		    set_nonblocking(sock, 0) <0) {
+			close(plainfd[0]);
+			close(plainfd[1]);
+			strncpy(errstr, "Cannot set socket options", errstr_len);
+			goto err;
+		}
+
+		ret = fork();
+		if (ret < 0)
+			err("Could not fork");
+		else if (ret == 0) {
+			// we are the child
+			signal (SIGPIPE, SIG_IGN);
+			close(plainfd[1]);
+			tlssession_mainloop(sock, plainfd[0], s);
+			close(sock);
+			close(plainfd[0]);
+			exit(0);
+		}
+		close(plainfd[0]);
+		close(sock);
+		sock = plainfd[1]; /* use the decrypted FD from now on */
+	}
+#else
+	if (keyfile) {
+		strncpy(errstr, "TLS requested but support not compiled in", errstr_len);
+		goto err;
+	}
+#endif
 	/* magic */
 	tmp64 = htonll(opts_magic);
 	WRITE_ALL_ERRCHK(sock, &tmp64, sizeof(tmp64), err,
@@ -430,7 +569,7 @@ end:
 }
 
 int setup_unix_connection(gchar * unixsock, gchar * name, CONNECTION_TYPE ctype,
-			  int *serverflags)
+			  int *serverflags, int testflags)
 {
 	struct sockaddr_un addr;
 	int sock;
@@ -452,7 +591,7 @@ int setup_unix_connection(gchar * unixsock, gchar * name, CONNECTION_TYPE ctype,
 		strncpy(errstr, strerror(errno), errstr_len);
 		goto err_open;
 	}
-	sock = setup_connection_common(sock, name, ctype, serverflags);
+	sock = setup_connection_common(sock, name, ctype, serverflags, testflags);
 	goto end;
 err_open:
 	close(sock);
@@ -463,7 +602,7 @@ end:
 }
 
 int setup_inet_connection(gchar * hostname, int port, gchar * name,
-			  CONNECTION_TYPE ctype, int *serverflags)
+			  CONNECTION_TYPE ctype, int *serverflags, int testflags)
 {
 	int sock;
 	struct hostent *host;
@@ -488,7 +627,7 @@ int setup_inet_connection(gchar * hostname, int port, gchar * name,
 		strncpy(errstr, strerror(errno), errstr_len);
 		goto err_open;
 	}
-	sock = setup_connection_common(sock, name, ctype, serverflags);
+	sock = setup_connection_common(sock, name, ctype, serverflags, testflags);
 	goto end;
 err_open:
 	close(sock);
@@ -499,14 +638,14 @@ end:
 }
 
 int setup_connection(gchar * hostname, gchar * unixsock, int port, gchar * name,
-		     CONNECTION_TYPE ctype, int *serverflags)
+		     CONNECTION_TYPE ctype, int *serverflags, int testflags)
 {
 	if (hostname != NULL) {
 		return setup_inet_connection(hostname, port, name, ctype,
-					     serverflags);
+					     serverflags, testflags);
 	} else if (unixsock != NULL) {
 		return setup_unix_connection(unixsock, name, ctype,
-					     serverflags);
+					     serverflags, testflags);
 	} else {
 		g_error("need a hostname or a unix domain socket!");
 		return -1;
@@ -602,7 +741,7 @@ int oversize_test(gchar * hostname, gchar * unixsock, int port, char *name,
 		if ((sock =
 		     setup_connection(hostname, unixsock, port, name,
 				      CONNECTION_TYPE_FULL,
-				      &serverflags)) < 0) {
+				      &serverflags, testflags)) < 0) {
 			g_warning("Could not open socket: %s", errstr);
 			retval = -1;
 			goto err;
@@ -670,6 +809,92 @@ err:
 	return retval;
 }
 
+int handshake_test(gchar * hostname, gchar * unixsock, int port, char *name,
+		   int sock, char sock_is_open, char close_sock, int testflags)
+{
+	int retval = -1;
+	int i = 0;
+	int serverflags = 0;
+	char buf[256];
+	u64 tmp64;
+	uint64_t mymagic = (name ? opts_magic : cliserv_magic);
+	uint32_t tmp32 = 0;
+
+	/* This should work */
+	if (!sock_is_open) {
+		if ((sock =
+		     setup_connection(hostname, unixsock, port, name,
+				      CONNECTION_TYPE_FULL,
+				      &serverflags, testflags)) < 0) {
+			g_warning("Could not open socket: %s", errstr);
+			goto err;
+		}
+	}
+
+	/* Intentionally throw an unknown option at the server */
+	tmp64 = htonll(opts_magic);
+	WRITE_ALL_ERRCHK(sock, &tmp64, sizeof(tmp64), err,
+			 "Could not write magic: %s", strerror(errno));
+	tmp32 = htonl(0x7654321);
+	WRITE_ALL_ERRCHK(sock, &tmp32, sizeof(tmp32), err,
+			 "Could not write option: %s", strerror(errno));
+	tmp32 = htonl((uint32_t) sizeof(tmp32));
+	WRITE_ALL_ERRCHK(sock, &tmp32, sizeof(tmp32), err,
+			 "Could not write option length: %s", strerror(errno));
+	WRITE_ALL_ERRCHK(sock, &tmp32, sizeof(tmp32), err,
+			 "Could not write option payload: %s", strerror(errno));
+	/* Expect proper error from server */
+	READ_ALL_ERRCHK(sock, &tmp64, sizeof(tmp64), err,
+			"Could not read magic: %s", strerror(errno));
+	tmp64 = ntohll(tmp64);
+	if (tmp64 != 0x3e889045565a9LL) {
+		strncpy(errstr, "magic does not match", errstr_len);
+		goto err;
+	}
+	READ_ALL_ERRCHK(sock, &tmp32, sizeof(tmp32), err,
+			"Could not read option: %s", strerror(errno));
+	tmp32 = ntohl(tmp32);
+	if (tmp32 != 0x7654321) {
+		strncpy(errstr, "option does not match", errstr_len);
+		goto err;
+	}
+	READ_ALL_ERRCHK(sock, &tmp32, sizeof(tmp32), err,
+			"Could not read status: %s", strerror(errno));
+	tmp32 = ntohl(tmp32);
+	if (tmp32 != NBD_REP_ERR_UNSUP) {
+		strncpy(errstr, "status does not match", errstr_len);
+		goto err;
+	}
+	READ_ALL_ERRCHK(sock, &tmp32, sizeof(tmp32), err,
+			"Could not read length: %s", strerror(errno));
+	tmp32 = ntohl(tmp32);
+	while (tmp32) {
+		char buf[1024];
+		size_t len = tmp32 < sizeof(buf) ? tmp32 : sizeof(buf);
+		READ_ALL_ERRCHK(sock, buf, len, err,
+				"Could not read payload: %s", strerror(errno));
+		tmp32 -= len;
+	}
+
+
+	/* Send NBD_OPT_ABORT to close the connection */
+	tmp64 = htonll(opts_magic);
+	WRITE_ALL_ERRCHK(sock, &tmp64, sizeof(tmp64), err,
+			 "Could not write magic: %s", strerror(errno));
+	tmp32 = htonl(NBD_OPT_ABORT);
+	WRITE_ALL_ERRCHK(sock, &tmp32, sizeof(tmp32), err,
+			 "Could not write option: %s", strerror(errno));
+	tmp32 = htonl((uint32_t) 0);
+	WRITE_ALL_ERRCHK(sock, &tmp32, sizeof(tmp32), err,
+			 "Could not write option length: %s", strerror(errno));
+
+	retval = 0;
+
+	g_message("Handshake test completed. No errors encountered.");
+err:
+	return retval;
+}
+
 int throughput_test(gchar * hostname, gchar * unixsock, int port, char *name,
 		    int sock, char sock_is_open, char close_sock, int testflags)
 {
@@ -698,7 +923,7 @@ int throughput_test(gchar * hostname, gchar * unixsock, int port, char *name,
 		if ((sock =
 		     setup_connection(hostname, unixsock, port, name,
 				      CONNECTION_TYPE_FULL,
-				      &serverflags)) < 0) {
+				      &serverflags, testflags)) < 0) {
 			g_warning("Could not open socket: %s", errstr);
 			retval = -1;
 			goto err;
@@ -751,10 +976,12 @@ int throughput_test(gchar * hostname, gchar * unixsock, int port, char *name,
 				req.type = htonl(NBD_CMD_FLUSH);
 				memcpy(&(req.handle), &j, sizeof(j));
 				req.from = 0;
+				req.len = 0;
 				if (write_all(sock, &req, sizeof(req)) < 0) {
 					retval = -1;
 					goto err_open;
 				}
+				req.len = htonl(1024);
 				++requests;
 			}
 		}
@@ -970,7 +1197,7 @@ int integrity_test(gchar * hostname, gchar * unixsock, int port, char *name,
 		if ((sock =
 		     setup_connection(hostname, unixsock, port, name,
 				      CONNECTION_TYPE_FULL,
-				      &serverflags)) < 0) {
+				      &serverflags, testflags)) < 0) {
 			g_warning("Could not open socket: %s", errstr);
 			goto err;
 		}
@@ -1495,6 +1722,10 @@ int main(int argc, char **argv)
 	int testflags = 0;
 	testfunc test = throughput_test;
 
+#if HAVE_GNUTLS
+	tlssession_init();
+#endif
+
 	/* Ignore SIGPIPE as we want to pick up the error from write() */
 	signal(SIGPIPE, SIG_IGN);
 
@@ -1511,7 +1742,7 @@ int main(int argc, char **argv)
 		exit(EXIT_FAILURE);
 	}
 	logging(MY_NAME);
-	while ((c = getopt(argc, argv, "FN:t:owfilu:")) >= 0) {
+	while ((c = getopt(argc, argv, "FN:t:owfilu:hC:K:A:H:")) >= 0) {
 		switch (c) {
 		case 1:
 			handle_nonopt(optarg, &hostname, &p);
@@ -1546,6 +1777,32 @@ int main(int argc, char **argv)
 		case 'u':
 			unixsock = g_strdup(optarg);
 			break;
+		case 'h':
+			test = handshake_test;
+			testflags |= TEST_HANDSHAKE;
+			break;
+#if HAVE_GNUTLS
+		case 'C':
+			certfile=g_strdup(optarg);
+			break;
+		case 'K':
+			keyfile=g_strdup(optarg);
+			break;
+		case 'A':
+			cacertfile=g_strdup(optarg);
+			break;
+		case 'H':
+			tlshostname=g_strdup(optarg);
+			break;
+#else
+		case 'C':
+		case 'K':
+		case 'H':
+		case 'A':
+			g_warning("TLS support not compiled in");
+			/* Do not change this - looked for by test suite */
+			exit(77);
+#endif
 		}
 	}
 
@@ -1553,6 +1810,12 @@ int main(int argc, char **argv)
 		handle_nonopt(argv[optind++], &hostname, &p);
 	}
 
+	if (keyfile && !certfile)
+		certfile = g_strdup(keyfile);
+
+	if (!tlshostname && hostname)
+		tlshostname = g_strdup(hostname);
+
 	if (test(hostname, unixsock, (int)p, name, sock, FALSE, TRUE, testflags)
 	    < 0) {
 		g_warning("Could not run test: %s", errstr);
